#include <vector>
#include <stdlib.h> /* rand() */
#include <time.h>

#define BISQBOT_INTERNAL
#include "bisqbot.hh"
static const unsigned MAX_FRAMES = 120;

extern "C" unsigned char RAM[0x800];
extern "C" void BotFrontMsg(const char*fmt, ...);

static unsigned PrevMarioX = 0, NFrames=0;
static unsigned MarioX     = 0;
static int MarioMovement;

void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y)
{
    /* GAME MONITORING */
    /* A, X, Y, PC correspond to the values of the CPU registers. */
    if(PC == 0x8082)
    {
        unsigned MarioXPos = RAM[0x86];
        unsigned MarioYPos = RAM[0xCE];
        unsigned ScrollBegin = RAM[0x73F];
        
        unsigned MarioScreenX = (unsigned char)(MarioXPos-ScrollBegin+256);
        
        MarioX = MarioXPos;
        if(NFrames++ > 0)
            MarioMovement += (signed char)(MarioX - PrevMarioX);
        PrevMarioX = MarioX;
    }
}

void BisqBotReset()
{
    fprintf(stderr, "restart\n");
    
    PrevMarioX = 0;
    NFrames = 0;
    MarioX = 0;
    MarioMovement = 0;
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    static const unsigned char InputTable[] =
    {
        K_R,
        K_R | K_A,
        K_L,
        K_L | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_L,
        K_B | K_L | K_A,

        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        K_B | K_R,
        K_B | K_R | K_A,
        
        K_B | K_R | K_L,
        K_R | K_L
    };

    static unsigned LastInput = 0x00;
    unsigned char Input = 0x00;
    
    if(!(rand()%20)) LastInput = InputTable[rand()%sizeof(InputTable)];
    
    Input |= LastInput;
    
    return Input;
}

unsigned BisqBotEvaluate(unsigned FrameNo)
{
    return MarioMovement + 512;
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo)
{
    if(FrameNo > MAX_FRAMES) return BisqBotRestart;
    fprintf(stderr, "%u: %d\n", FrameNo, MarioMovement);
    return BisqBotContinue;
}
