#include <vector>
#include <stdlib.h> /* rand() */
#include <time.h>

#define BISQBOT_INTERNAL
#include "bisqbot.hh"
static const unsigned MAX_FRAMES = 120;

static BisqBotVar Bonus, Hurt, Item, Kill, BossHit;
static BisqBotVar PauseActivated, Scrolling;
static unsigned LastTick;

/* Scrolling:
    0,1,2,3 = r,u,l,d
    5=prev, 6=next
*/

extern "C" unsigned char RAM[0x800];
extern "C" void BotFrontMsg(const char*fmt, ...);

static bool InSoundCode=false;
    
void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y)
{
    /* GAME MONITORING */
    /* A, X, Y, PC correspond to the values of the CPU registers. */
    
    if(PC == 0xD551) { InSoundCode=true; return; }
    if(PC == 0xD554) { InSoundCode=false; return; }
    if(InSoundCode) return;
    
    static const unsigned short  FrameUpd   = 0xD57D;
    static const unsigned short  RandomFunc = 0xC5AB;
    static const unsigned short  CreateDrop = 0xBF25;
    static const unsigned short  HurtFunc   = 0xA244;
    static const unsigned short  GotItem    = 0xC833;
    static const unsigned short  KillFunc   = 0xC219;
    static const unsigned short  EnemyKill  = 0xBEEC;
    static const unsigned short  BossDamaged= 0xC90F;
    static const unsigned short  BossKilled = 0xC916;
    static const unsigned short SelectPushed= 0x91D1;
    static const unsigned short PauseRelease= 0x91F5;
    static const unsigned short  EnemyCreate= 0xF663;
    static const unsigned short  HorizScroll= 0x9FE8;
    static const unsigned short   VertScroll= 0x92A6;
    static const unsigned short   PrevScroll= 0x925B;
    static const unsigned short   NextScroll= 0x92A6;
    
    unsigned char GameFrame = RAM[0x23];
    unsigned char Bank      = RAM[0x42];
    
    #define TestPC(rule) \
        if(rule) BotFrontMsg("[%u,%02X,%02X]Match %s (%04X,%02X,%02X,%02X)\n", \
            LastTick, GameFrame, Bank, \
            #rule, \
            PC,A,X,Y); \
        if(rule)
    
#if 0
    // Display Rockman's Y pos
    if(PC == FrameUpd) fprintf(stderr, "%02X ", RAM[0x600]);
#endif
    
    TestPC(PC == CreateDrop)  Bonus.Set(Y);
    TestPC(PC == HurtFunc)    Hurt.Set(1);
    TestPC(PC == KillFunc)    Hurt.Set(2);
    TestPC(PC == GotItem)     Item.Set(A);
    TestPC(PC == EnemyKill)   Kill.Set(A);
    TestPC(PC == BossDamaged) BossHit.Set(1);
    TestPC(PC == BossKilled)  BossHit.Set(2);
    TestPC(PC == SelectPushed && Bank == 0x05)
    {
        PauseActivated.Set(1);
    }
    TestPC(PC == PauseRelease && Bank == 0x05)
    {
    }
    TestPC(PC == EnemyCreate);
    TestPC(PC == HorizScroll && Bank == 0x05)
    {
        Scrolling.Set(RAM[0x26] & 3);
    }
    TestPC(PC == VertScroll && Bank == 0x05)
    {
        Scrolling.Set(RAM[0x26] & 3);
    }
    TestPC(PC == PrevScroll && Bank == 0x05)
    {
        Scrolling.Set(5);
    }
    TestPC(PC == NextScroll && Bank == 0x05)
    {
        Scrolling.Set(6);
    }
}

static bool PauseActive = false;
static bool WantPause = false;
static int UnPause=0;
static int Climb=0;

void BisqBotReset()
{
    Bonus.Clear();
    Hurt.Clear();
    Item.Clear();
    Kill.Clear();
    BossHit.Clear();
    
    PauseActivated.Clear();
    Scrolling.Clear();
    
    WantPause   = false;
    PauseActive = false;
    Climb       = 0;
    UnPause=0;
    
    fprintf(stderr, "restart\n");
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    static const unsigned char InputTable[] =
    {
        K_L,
        K_L | K_A,
        K_L | K_U | K_A,
        K_R,
        K_R | K_A,
        K_R | K_U | K_A,
        
        K_A, K_A, K_A, K_A, K_A,
        K_U | K_A
    };

    static const unsigned ClimbFlags = K_U | K_D;
    static unsigned LastInput = 0x00;
    unsigned char Input = 0x00;

//    if(!(rand()%40)) Input |= K_B;
//    if(!(rand()%20)) Input |= K_ST;
//    if(!(rand()%30)) Input |= K_SE;

//    if(!(rand()%20)) LastInput ^= K_L;
//    if(!(rand()%20)) LastInput ^= K_R;
//    if(!(rand()%20)) LastInput ^= K_D;
//    if(!(rand()%20)) LastInput ^= K_U;
//    if(!(rand()%30)) LastInput ^= K_A;
    
//    if(!rand()%30)) LastInput ^= K_U;
    
    if(!(rand()%20)) LastInput = InputTable[rand()%sizeof(InputTable)];
    
//    LastInput |= K_A;
//    LastInput |= K_L;

    Input |= LastInput;
    
    /*
    if(Climb > 0)
    {
        Input = Climb;
    }
    else if(Input & ClimbFlags)
    {
        Climb = Input & ClimbFlags;
    }
    */
    
    if(WantPause)
    {
        Input |= K_SE; //select button
    }
    else if(PauseActive)
    {
        if(FrameNo == UnPause)
        {
            Input |= K_SE;
            PauseActive=false;
            UnPause=0;
        }
    }
    
    return Input;
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo)
{
    LastTick=FrameNo;
    if(BossHit.Has(1) && !BossHit.Has(2))
    {
        BossHit.Clear();
        WantPause = true;
    }
    if(PauseActivated.Count() > 0)
    {
        WantPause=false;
        PauseActivated.Clear();
        if(!PauseActive)
        {
            PauseActive=true;
            UnPause = FrameNo + 2;
            srand(time(NULL));
        }
    }

/**/
    if(Hurt.Count())
    {
        Hurt.Clear();
        WantPause=true;
    }
/**/

/**/
    if(RAM[0x400] == 0x06)
    {
        fprintf(stderr, "Standing at %02X\n", RAM[0x600]);
        if(RAM[0x600] > 0x74) return BisqBotRestart;
        return BisqBotAccept;
    }
    if(FrameNo > MAX_FRAMES) return BisqBotRestart;
    return BisqBotContinue;
/**/

    if(Hurt.Has(2)) srand(time(NULL));
    
    if(Hurt.Count()
    || Item.Count()
    //|| RAM[0x600] > 0x64
    || !Bonus.LimitedTo(0x3B, 0x3F)
    || FrameNo > MAX_FRAMES)
    {
        return BisqBotRestart;
    }
    
    //if(BossHit.Has(2))
    //if(Bonus.Has(0x3F, 1))
    //if(Kill.Count())
    if(Scrolling.Has(6))
    {
        return BisqBotAccept;
    }
    return BisqBotContinue;
}
