#include <cmath>
#include <cstring>
#include <algorithm>

#include <vector>
#include <map>

#include <complex>

#include "beauty.hh"

#include "../../types.h"
extern "C" {
#include "../../state.h"
#include "../../movie.h"
#include "../../driver.h"
#include "../../video.h"
}

#include "../../coroutine.h"

typedef std::complex<double> complex;

int do_beauty_analysis = 0;

struct SPR
{
        unsigned char y,no,atr,x;
        bool operator< (const SPR& b) const
        {
            if(y != b.y) return y < b.y;
            if(x != b.y) return x < b.x;
            if(atr != b.atr) return atr < b.atr;
            return no < b.no;
        }
        bool operator== (const SPR& b) const
        {
            return y == b.y && no == b.no && atr == b.atr && x == b.x;
        }
};
extern "C" {
extern SPR SPRAM[64]; /* actually uint8[256] */
}

/* Analyzes beauty in the given NES screenshot.
 */

static const unsigned char* xbuf = 0;
static unsigned width;

/* Calculate average tint in given section of the screen. */
static complex GetAverageTint(int x0,int y0, int x1,int y1)
{
    double avg_i = 0, avg_q = 0, avg_y = 0;
    double count = 0;

    for(int y=y0; y<=y1; ++y)
        for(int x=x0; x<=x1; ++x)
        {
            unsigned char pix = xbuf[x + y * width];
            
            int color = pix & 0x0F;
            /* phases [i] = cos( i * PI / 6 ) */

            static double const lo_levels [4] = { -0.12f, 0.00f, 0.31f, 0.72f };
            static double const hi_levels [4] = {  0.40f, 0.68f, 1.00f, 1.00f };
            static double const phases [0x10 + 3] = {
                -1.0f, -0.866025f, -0.5f, 0.0f,  0.5f,  0.866025f,
                 1.0f,  0.866025f,  0.5f, 0.0f, -0.5f, -0.866025f,
                -1.0f, -0.866025f, -0.5f, 0.0f,  0.5f,  0.866025f,
                 1.0f
            };
            #define TO_ANGLE_SIN( color )   phases [color]
            #define TO_ANGLE_COS( color )   phases [(color) + 3]
            int level = (pix >> 4) & 0x03;
            double lo = lo_levels[level], hi = hi_levels[level];
            if(color==0) lo=hi;
            else if(color==0x0D)hi=lo;
            else if(color>0x0D) hi=lo=0;
            
            double sat = (hi - lo) * 0.5f;
            double i = TO_ANGLE_SIN( color ) * sat;
            double q = TO_ANGLE_COS( color ) * sat;
            double y = (hi + lo) * 0.5f;
            
            double pow = 1;
            
            avg_i += i*y;
            avg_q += q*y;
            avg_y += y;
            
            count += pow;
        }

    avg_i /= count;
    avg_q /= count;
    avg_y /= count;
    
    complex res(avg_i, avg_q); res *= 100;
    
    double y = std::abs(res);
    if(y >= 0.0) res *= std::log(1 + y) / (1 + y);
    return res;
    
    //return complex ( atan2(avg_i, avg_q) , std::log(1 + avg_y));
}

extern "C" {
    extern uint32_t framecount;
    
    void GetScroll(unsigned& First, unsigned& Last);
    extern unsigned VScroll,HScroll;
    extern unsigned ScrollPointers[272];
}

struct region
{
    int x0,x1, y0,y1;
    
    complex tint;
    
    int sprite_jitter;
};
static double CalculateBeauty(unsigned w,unsigned h)
{
    std::vector<region> regions;
    regions.reserve(4*3);
    for(unsigned x=0; x<4; ++x)
        for(unsigned y=0; y<3; ++y)
        {
            region tmp;
            tmp.tint=0;
            tmp.sprite_jitter=0;
            tmp.x0 = x*w/4;
            tmp.x1 = (x+1)*w/4;
            tmp.y0 = y*h/3;
            tmp.y1 = (y+1)*h/3;
            regions.push_back(tmp);
        }
/*    
     regions[5] =
    {
        // corners: nw, ne, sw, se, middle
        { w*0 /99,w*50/99, h*0 /99,h*50/99, 0, 0 },
        { w*50/99,w*99/99, h*0 /99,h*50/99, 0, 0 },
        { w*0 /99,w*50/99, h*50/99,h*99/99, 0, 0 },
        { w*50/99,w*99/99, h*50/99,h*99/99, 0, 0 },
        { w*33/99,w*55/99, h*33/99,h*55/99, 0, 0 }
    };
*/
    unsigned FirstLine=0, LastLine=h-1;
    GetScroll(FirstLine,LastLine);
    
    static SPR old_sprites[64];
    SPR new_sprites[64];
    std::memcpy(&new_sprites, SPRAM, sizeof(new_sprites));
    
    std::sort(new_sprites, new_sprites+64);
    
    for(unsigned b=0; b<64; ++b)
    {
        SPR& spr = new_sprites[b];
        SPR& old = old_sprites[b];
        
        /* If the sprite appears to have moved, compensate for scrolling and try again */
        if(spr.x != old.x || spr.y != old.y)
        {
            spr.x -= ScrollPointers[spr.y];
            spr.y -= VScroll;
        }
        
        if(std::memcmp(&spr, &old, sizeof(spr)) != 0)
        {
            old = spr;
            
            if(spr.y >= 224) continue;
            
            for(unsigned a=0; a< regions.size(); ++a)
            {
                /* Check if this sprite affects this region */
                region& reg = regions[a];
                
                if(spr.x+8 >= reg.x0 && spr.x < reg.x1
                && spr.y+8 >= reg.y0 && spr.y < reg.y1)
                {
                    reg.sprite_jitter += 1;
                    if(spr.x != old.x || spr.y != old.y)
                        reg.sprite_jitter += 2;
                }
            }
        }
    }

    double totaltint = 0;
    for(unsigned a=0; a< regions.size(); ++a)
    {
        region& reg = regions[a];
        reg.tint = GetAverageTint(reg.x0, reg.y0, reg.x1, reg.y1);
        totaltint += std::abs(reg.tint);
    }
    totaltint /= regions.size();
    
    /*
    PLAN
    
    Analyze five sections of this screen.
    
    - Each four quarters of the screen
    - And the middle (slightly larger than a quarter?)
    
    Analyze:
    - Average tint
    - Amount of motion (ignore global motion)
    
    Value:
    - Non-bleak tints
    - Different tint in each section of screen
    - Motion appears in at least two different sections of screen
    
    */
    
    if(do_beauty_analysis >= 3)
    {
        fprintf(stderr, "analysis: ");
        for(unsigned a=0; a< regions.size(); ++a)
        {
            region& reg = regions[a];
            fprintf(stderr, "(%3.1f<%+3.1f)(%2d) ",
                std::abs(reg.tint),
                std::arg(reg.tint),
                reg.sprite_jitter);
        }
        fprintf(stderr, "\n");
    }

/*
    complex t = regions[0].tint + regions[1].tint + regions[2].tint + regions[3].tint;
    t /= 4;
*/

    double interestingness = 0;
    
    if(totaltint > 0.1)
    {
        for(unsigned a=0; a<regions.size(); ++a)
            for(unsigned b=a+1; b<regions.size(); ++b)
            {
                double mul = 1;
                if(regions[a].x0 != regions[b].x0) mul *= 2;
                if(regions[a].y0 != regions[b].y0) mul *= 2;

                double tintdiff = std::abs(regions[a].tint - regions[b].tint) * 0.003;
                interestingness += tintdiff * mul;
                
                interestingness += std::log(2 + regions[a].sprite_jitter / 4.0)
                                 * std::log(2 + regions[b].sprite_jitter / 4.0)
                                 * mul * 100;
            }
    }
/*
        std::abs(regions[0].tint - regions[1].tint)
      + std::abs(regions[0].tint - regions[2].tint)
      + std::abs(regions[0].tint - regions[3].tint)
      + std::abs(regions[1].tint - regions[2].tint)
      + std::abs(regions[1].tint - regions[3].tint)
      + std::abs(regions[2].tint - regions[3].tint)
      + std::abs(regions[4].tint - t)
      +  (std::log(1 + regions[0].sprite_jitter / 8.0)
        * std::log(1 + regions[3].sprite_jitter / 8.0)
      +   std::log(1 + regions[1].sprite_jitter / 8.0)
        * std::log(1 + regions[2].sprite_jitter / 8.0)
      +   std::log(1 + regions[3].sprite_jitter / 4.0)
         );
*/
    return interestingness;
}

struct SaveState
{
    std::vector<unsigned char> Data;

    void Create()
    {
        FILE*fp = tmpfile();
        FCEUSS_SaveFP(fp);
        rewind(fp);
        
        fseek(fp,0,SEEK_END);
        long pos = ftell(fp);
        rewind(fp);
        
        SaveState result;
        Data.resize(pos);
        fread(&Data[0], 1, pos, fp);
        fclose(fp);
    }
    void Load()
    {
        FILE*fp = tmpfile();
        fwrite(&Data[0], 1, Data.size(), fp);
        rewind(fp);
        FCEUSS_LoadFP(fp, 0);
        fclose(fp);
    }
};

static std::vector<double> interestingness_map;
static std::map<int, SaveState> some_savestates;
static std::vector<unsigned> review_results;

static bool LocateBeauty(unsigned frameno)
{
    scrBegin;
    if(true) /* scope */
    {
        std::map<int, SaveState>::iterator
            i = some_savestates.lower_bound(frameno);

        for(;;)
        {
            if(i != some_savestates.end())
            {
                if(i->first <= frameno)
                {
                    if(framecount > frameno || i->first > framecount)
                        i->second.Load();
                    
                    fprintf(stderr, "Aiming for frame %u, loading one that gives %u hopefully (got %u)\n",
                        frameno, i->first, framecount);
                    break;
                }
            }
            if(i == some_savestates.begin()) break;
            --i;
        }
    }
    
    while(framecount < frameno)
    {
        //fprintf(stderr, "Frame %u\n", framecount);
        FCEUI_FrameAdvance();
        scrReturn(false);
    }
    scrFinish(true);
}

extern "C" void CloseStuff(int signum);

static void CaptureBeauty()
{
    scrBegin;
    static unsigned a=0;
    while(a < review_results.size())
    {
        fprintf(stderr, "Capturing beauty %u at frame %u\n", a, review_results[a]);
        
        while(!LocateBeauty(review_results[a]))
        {
            scrReturnV;
        }

        /* capture it */

        char beauty_fn[64];
        sprintf(beauty_fn, "beauty_%u.png", a);
        fprintf(stderr, "- Capturing %s, have frame %u now\n", beauty_fn, framecount);
        
        SaveSnapshot(beauty_fn);

        ++a;
    }
    
    /* all beauties have been analyzed, end stuff now */
    
    CloseStuff(0);
    scrFinishV;
}

struct section
  { unsigned sect; double value;
    bool operator< (const section& sect) const
    { return value > sect.value; }
    bool operator== (const section& sect) const
    { return value == sect.value; }
  };


static void ReviewBeauty()
{
    const int num_seconds = framecount / 60;
    const unsigned num_sections = std::max(
      std::min(50, num_seconds),
        // not more than 50
        // not more than one per each second of movie
      num_seconds / 10
        // but still at least one per each 10 seconds of movie
    );
    const unsigned num_results  = std::min(
      std::min(20, (int)( num_seconds / (60/5) ) ),
         // not more than 20
         // not more than 5 per each minute of movie
         // not more than the number of sections
      (int)(num_sections));
    
    std::vector<section> sects(num_sections);
    
    for(unsigned a=0; a<num_sections; ++a)
        sects[a].sect = a;
    
    for(unsigned a=0; a<interestingness_map.size(); ++a)
    {
        const double val = interestingness_map[a];
        const unsigned sect = a * num_sections / interestingness_map.size();
        if(val > sects[sect].value) sects[sect].value = val;
    }
    
    std::sort(sects.begin(), sects.end());
    
    for(unsigned a=0; a<num_results; ++a)
    {
        unsigned sect = sects[a].sect;
        unsigned sect_begin = (sect) * interestingness_map.size() / num_sections;
        unsigned sect_end = (sect+1) * interestingness_map.size() / num_sections;
        
        for(unsigned b=sect_begin; b<sect_end; ++b)
            if(interestingness_map[b] == sects[a].value)
            {
                fprintf(stderr, "Selected beauty at frame %u (%g)\n", b, sects[a].value);
                review_results.push_back(b);
                break;
            }
    }
    std::sort(review_results.begin(), review_results.end());
}

void Beauty(const unsigned char* Xbuf, unsigned w, unsigned h)
{
    if(!do_beauty_analysis) return;
    
    xbuf = Xbuf;
    width = w;
    
    static bool finished_recording = false;
    
    if(!finished_recording) /* Record interestingness */
    {
        double interestingness = CalculateBeauty(w,h);

        if(do_beauty_analysis >= 2)
            std::printf(
                "Beauty %u: %g\n", framecount, interestingness);

        if(interestingness_map.size() <= framecount)
            interestingness_map.resize(framecount+1);
        interestingness_map[framecount] = interestingness;
        
        if(framecount == 1 || !(framecount % 100))
        {
            some_savestates[framecount].Create();
        }
        
        static unsigned maxframe = ~0;
        if(maxframe == (unsigned)~0)
        {
            const char* env = getenv("MAXFRAMES");
            maxframe = env ? atoi(env) : 0;
        }

        finished_recording = framecount == maxframe-2;
        
        if(!finished_recording)
        {
            return;
        }
    }
    
    /* Analyze interestingness */
    if(review_results.empty())
        ReviewBeauty();

    CaptureBeauty();
}
